--- SmartRes2
-- @class file
-- @name SmartRes2.lua
-- @author Myrroddin of Llane

-- localise global variables for faster access ------------------------------
local _G = getfenv(0)

local string = _G.string

local table = _G.table
local tinsert = table.insert
local tsort = table.sort
local wipe = table.wipe

local pairs = _G.pairs
local ipairs = _G.ipairs

-- Upvalued Blizzard API ----------------------------------------------------
local GetNumRaidMembers = _G.GetNumRaidMembers
local GetNumPartyMembers = _G.GetNumPartyMembers
local GetSpellInfo = _G.GetSpellInfo
local IsSpellInRange = _G.IsSpellInRange
local UnitClass = _G.UnitClass
local UnitInRaid = _G.UnitInRaid
local UnitInRange = _G.UnitInRange
local UnitIsDead = _G.UnitIsDead
local UnitIsGhost = _G.UnitIsGhost
local UnitLevel = _G.UnitLevel

-- declare addon ------------------------------------------------------------
local LibStub = _G.LibStub

local SmartRes2 = LibStub("AceAddon-3.0"):NewAddon("SmartRes2", "AceConsole-3.0", "AceEvent-3.0", "LibBars-1.0")
local L = LibStub("AceLocale-3.0"):GetLocale("SmartRes2", true)

-- get version from .toc - set to Development if no version
local version = GetAddOnMetadata("SmartRes2", "Version")

-- add localisation to addon
SmartRes2.L = L
-- declare the database
local db

-- debugging section --------------------------------------------------------

--[===[@debug@
_G.SmartRes2 = SmartRes2
--@end-debug@]===]

-- additional libraries -----------------------------------------------------

-- LibDataBroker used for LDB enabled addons like ChocolateBars
local DataBroker = LibStub:GetLibrary("LibDataBroker-1.1", true)
if not DataBroker then
	LoadAddOn("LibDataBroker-1.1")
	local DataBroker = LibStub("LibDataBroker-1.1", true)
	if not DataBroker then
		error(format(L["%s requires the library '%s' to be available."], "SmartRes2", "LibDataBroker-1.1"))
	end
end
-- LibBars used for bars
local Bars = LibStub:GetLibrary("LibBars-1.0", true)
if not Bars then
	LoadAddOn("LibBars-1.0")
	local Bars = LibStub:GetLibrary("LibBars-1.0", true)
	if not Bars then
		error(format(L["%s requires the library '%s' to be available."], "SmartRes2", "LibBars-1.0"))
	end
end
-- LibResComm used for communication
local ResComm = LibStub:GetLibrary("LibResComm-1.0", true)
if not ResComm then
	LoadAddOn("LibResComm-1.0")
	local ResComm = LibStub:GetLibrary("LibResComm-1.0", true)
	if not ResComm then
		error(format(L["%s requires the library '%s' to be available."], "SmartRes2", "LibResComm-1.0"))
	end
end
-- LibSharedMedia used for more textures
local Media = LibStub:GetLibrary("LibSharedMedia-3.0", true)
-- register the res bar textures with LibSharedMedia-3.0
Media:Register("statusbar", "Blizzard", [[Interface\TargetingFrame\UI-StatusBar]])

-- global variables ---------------------------------------------------------

-- local variables ----------------------------------------------------------
local doingRessing = {}
local waitingForAccept = {}
local resBars = {}
local orientation
local icon
local LastRes

--[===[@debug@
SmartRes2.doingRessing = doingRessing
SmartRes2.waitingForAccept = waitingForAccept
--@end-debug@]===]

-- variable to use for multiple PLAYER_REGEN_DISABLED calls (see SmartRes2:PLAYER_REGEN_DISABLED below)
local in_combat = false
local Player = UnitName("player")

-- addon defaults -----------------------------------------------------------

local defaults = {
	profile = {
		autoResKey = "",
		borderThickness = 10,
		chatOutput = "0-none",
		classColours = true,
		collisionBarsColour = { r = 1, g = 0, b = 0, a = 1 },
		disableAddon = false,
		hideAnchor = true,
		horizontalOrientation = "RIGHT",
		manualResKey = "",
		notifyCollision = "0-off",
		notifySelf = true,
		randMsgs = false,
		customchatmsg = "",
		resBarsColour = { r = 0, g = 1, b = 0, a = 1 },
		resBarsIcon = true,		
		resBarsAlpha = 1,
		resBarsBorder = nil,
		resBarsTexture = "Blizzard",
		resBarsX = 0,
		resBarsY = 600,
		reverseGrowth = false,
		scale = 1,
		showBattleRes = false,		
		visibleResBars = true,
		waitingBarsColour = { r = 0, g = 0, b = 1, a = 1 },
		randChatTbl = {
			[1] = L["%%p%% is bringing %%t%% back to life!"],
			[2] = L["Filthy peon! %%p%% has to resurrect %%t%%!"],
			[3] = L["%%p%% has to wake %%t%% from eternal slumber."],
			[4] = L["%%p%% is ending %%t%%'s dirt nap."],
			[5] = L["No fallen heroes! %%p%% needs %%t%% to march forward to victory!"],
			[6] = L["%%p%% doesn't think %%t%% is immortal, but after this res cast, it is close enough."],
			[7] = L["Sleeping on the job? %%p%% is disappointed in %%t%%."],
			[8] = L["%%p%% knew %%t%% couldn't stay out of the fire. *Sigh*"],
			[9] = L["Once again, %%p%% pulls %%t%% and their bacon out of the fire."],
			[10] = L["%%p%% thinks %%t%% should work on their Dodge skill."],
			[11] = L["%%p%% refuses to accept blame for %%t%%'s death, but kindly undoes the damage."],
			[12] = L["%%p%% grabs a stick. A-ha! %%t%% was only temporarily dead."],
			[13] = L["%%p%% is ressing %%t%%"],
			[14] = L["%%p%% knows %%t%% is faking. It was only a flesh wound!"],
			[15] = L["Oh. My. God. %%p%% has to breathe life back into %%t%% AGAIN?!?"],
			[16] = L["%%p%% knows that %%t%% dying was just an excuse to see another silly random res message."],
			[17] = L["Think that was bad? %%p%% proudly shows %%t%% the scar tissue caused by Hogger."],
			[18] = L["Just to be silly, %%p%% tickles %%t%% until they get back up."],
			[19] = L["FOR THE HORDE! FOR THE ALLIANCE! %%p%% thinks %%t%% should be more concerned about yelling FOR THE LICH KING! and prevents that from happening."],
			[20] = L["And you thought the Scourge looked bad. In about 10 seconds, %%p%% knows %%t%% will want a comb, some soap, and a mirror."],
			[21] = L["Somewhere, the Lich King is laughing at %%p%%, because he knows %%t%% will just die again eventually. More meat for the grinder!!"],
			[22] = L["%%p%% doesn't want the Lich King to get another soldier, so is bringing %%t%% back to life."],
			[23] = L["%%p%% wonders about these stupid res messages. %%t%% should just be happy to be alive."],
			[24] = L["%%p%% prays over the corpse of %%t%%, and a miracle happens!"],
			[25] = L["In a world of resurrection spells, why are NPC deaths permanent? It doesn't matter, since %%p%% is making sure %%t%%'s death isn't permanent."],
			[26] = L["%%p%% performs a series of lewd acts on %%t%%'s still warm corpse. Ew."],
			[27] = L["%%t%%, you owe me a cookie for this."],
			[28] = L["Fear not, %%t%%; my res is not 'bearform CPR.'"],
			[29] = L["Don't ask where I'm putting this acorn, %%t%%, but it's uncomfortable enough to wake the dead."],
			[30] = L["Rise up and fight once more, %%t%%! For honor, for glory, for a chicken sandwich!!"],
			[31] = L["%%t%%... I know they said there was candy over there, but you can't eat it if you're dead!"],
			[32] = L["%%p%%-Bot in operation. Scanning puddle of goo labeled '%%t%%.' Prognosis: lifesigns negative. Commencing resurrection."],
			[33] = L["*%%p%% pokes %%t%% with a stick* Um, guys? I think it's still alive..."],
			[34] = L["And the good thing is, %%t%%'s not dead for long, I GET TO KILL %%t%% AGAIN!"],
			[35] = L["So, how much experience do I get for killing %%t%%?"],
			[36] = L["Oi, I thought we had a rule about 'No dieing, %%t%%'!"],
			[37] = L["Once upon a time there was a little raider named %%t%%... and they loved to hug the floor. The end."],
			[38] = L["I think %%t%% is taking 'DIAF' a bit too literally, especially with all these fires around. Maybe we should change it to something like, Die In An Ocean. DIAO! Has a ring to it... if only the next expansion wasn't flooded with Naga. Oh well."],
			[39] = L["Killing you once is never enough for me, %%t%%. How...how gratifying to do it another time over."],
			[40] = L["%%t%%! What have I told you about running with scissors?!"],
			[41] = L["A wild %%t%% appears! I choose you, Pikachu! Thundershock-res, go!"],
			[42] = L["I choose you, %%t%%!"],
			[43] = L["A wild %%t%%! It looks weakened. Pokball, go!"],
			[44] = L["Dream as if you'll live forever, %%t%%... and live as if you'll die today and I decide not to res you again."],
			[45] = L["Look at my mask, %%t%%! Isn't it pretty? It raises the dead!"],
			[46] = L["Ah ha! %%t%% down! FLAMING HANDS OF FIREY DOOM--I mean... reviving."],
			[47] = L["You have to remember something: we're dealing with raiders like %%t%%. They need to be terrified of death. It's like mother's milk to them. Without it, their bones won't grow back properly when we res them."],
			[48] = L["Rebirthing now - I wanna live my life, wanna give you everything - Breathe for the first time now - %%t%% comes alive somehow"],
			[49] = L["Days go on forever - But I have not left your side - We can chase the dark together - If %%t%% goes then so will I"],
			[50] = L["Yay! I always wanted my very own %%t%%-zombie!"],
			[51] = L["Death-defying feats are clearly not your strong point, %%t%%."],
			[52] = L["Arise, %%t%%, and fear death no longer... or at least until the next pull."],
			[53] = L["%%t%%, this better not be another attempt to get me to give you mouth-to-mouth."],
			[54] = L["%%t%%, stop hitting on the Spirit Healer and come kill something!"],
			[55] = L["Cool, I received 42 silver, 32 copper from Corpse of %%t%%."],
			[56] = L["Resurrecting %%t%%. Side effects may include: nausea, explosive bowels, and a craving for brains. Resurrection is not for everyone. Please consult healer before dying."],
			[57] = L["GAME OVER, %%t%%. To continue press up, up, down, down, left, right, left, right, B, A, B, A, Select + Start"],
			[58] = L["%%t%%, your subscription to Life(c) has expired. Do you wish to renew?"],
			[59] = L["It just so happens that %%t%% is only MOSTLY dead. There's a big difference between mostly dead and all dead. Mostly dead is slightly alive."],
			[60] = L["Death comes for you, %%t%%! With large, pointy teeth!"],
			[61] = L["Wee-ooo! Wee-ooo! Wee-ooo! Wee-ooo! Wee-ooo! Wee-ooo... that's the best ambulance impression I can do, %%t%%."],
			[62] = L["Hey %%t%%! Don't go towards the light! Well, unless it says 'Accept' ... but even then, it might be a trick!"],
			[63] = L["Applying my magical life-giving spatula to %%t%%, the perfect tool for un-sticking a faceplant!"],
			[64] = L["The play is currently under review.... The ruling on the field has been overturned. %%t%% will not be charged a Time Out."],
			[65] = L["Well, %%t%%, if you had higher faction with (Player Name), you might have gotten a heal. How do you raise it? 1g donation= 15 rep."],
			[66] = L["Game Over, %%t%%. Insert 50g to play again."],
			[67] = L["Hope this works, %%t%%. The last one I tried this on couldn't stop eating everyone's brains."],
			[68] = L["%%t%% Dead. Me press button. It bring back %%t%%. We happy and hold hands :D"],
			[69] = L["Dammit, %%t%%, get up! How do you expect me to loot the mobs you're supposed to be killing?"],
			[70] = L["Sorry for the delay %%t%%, but I had to get your soul shard back from the warlock."],
			[71] = L["The first time you mention 'Braaaains' I'm putting you back down, %%t%%."],
			[72] = L["%%t%%, before I res you there is something you should know. I am your father's brother's nephew's cousin's former roommate."],
			[73] = L["Don't worry, %%t%%, tis but a scratch!"],
			[74] = L["How are you feeling, %%t%%? Sore? Everywhere? Need cookies?"],
			[75] = L["Gentlemen, we can rebuild him. We have the technology. We have the capability to build the world's first bionic man. %%t%% will be that man. Better than he was before. Better, stronger, faster."],
			[76] = L["Forgive me, %%t%%, your death only adds to my failure!"],
			[77] = L["Forgive me, %%t%%! Your death only adds to your repair bill!"],
			[78] = L["The play is currently under review.... The ruling on the field has been overturned. %%t%% will not be charged a Time Out."],
			[79] = L["studiesshowthatbythetimeyoufinishreadingallofthis %%t%% willbealive!"],
			[80] = L["Ok, one more time. The bad guys are supposed to die, the good guys are supposed to live. And %%t%%? We're the good guys."],
			[81] = L["%%t%%, stop worshiping the ground I walk on and get up."],
			[82] = L["It's not that I didn't try to heal you, %%t%%, I simply found it more mana efficient to rez you."],
			[83] = L["DISCLAIMER: I learned how to rez from a recipe on the AH, so accept at your own risk, %%t%%."],
			[84] = L["Now ressing %%t%%... step back, this could get ugly!"],
			[85] = L["Walk it off, %%t%%."],
			[86] = L["%%t%%'s death was merely a setback!"],
			[87] = L["Get up, %%t%%. It's not healthy to be dead that long."],
			[88] = L["Well, %%t%%, if you had higher faction with me, you might have gotten a heal. How do you raise it? 1g donation= 15 rep."],
			[89] = L["%%t%% has encountered a fatal error. Rebooting..."],
			[90] = L["Giving %%t%% a shot at another 10 percent durability loss, please stand by."],
			[91] = L["Let's give %%t%% a round of applause for that beautiful demonstration of what not to do."],
			[92] = L["Anybody know what the respawn time on %%t%% is?"],
			[93] = L["Sorry %%t%%, I was too busy rolling my face across the keyboard while shouting PEW PEW!"],
			[94] = L["Sorry %%t%%, it was either keep you alive or heat up some pizza rolls."],
			[95] = L["Well %%t%%, you tried your best. And apparently failed miserably. Good job."],
			[96] = L["Giving %%t%% another chance to noob it up."],
			[97] = L["I used to try mouth-to-mouth, but %%t%% always came back pretty pissed off."],
			[98] = L["Come on, %%t%%, you need to learn to heal yourself better, I have wanding to do."],
			[99] = L["Grandaddy always said laughter was the best medicine. I guess it wasn't strong enough to keep %%t%% alive."],
			[100] = L["Don't rush me %%t%%. You rush a miracle worker, you get rotten miracles."],
			[101] = L["%%p%%'s corpse removal service here to serve. Looks like you have a fresh %%t%% for me, I'll get that out of your hair in about 8.89 seconds."],
			[102] = L["Don't rez %%t%% yet! I'm still extracting a kidney..."],
			[103] = L["Repeat after me, %%t%%: 'Goblin landmines are NOT frizbees.'"],
			[104] = L["%%t%% 0, monster 1, and the crowd roars!!"],
			[105] = L["%%t%%, spamming 'beg' was a nice touch, but the monster was clearly unimpressed."],
			[106] = L["How's that ground taste, %%t%%?"],
			[107] = L["Congratulations %%t%%, you passed the idiot test. You are now certified!"],
			[108] = L["Casting Siphon Gold on %%t%%. Please press accept to rez and transfer 100 gold to %%p%%'s bank."],
			[109] = L["Today's category on Final Jeopardy: Spells that end in 'essurection'. %%t%%, ready?"],
			[110] = L["The name of the game is 'Stay Alive'. %%t%% fails. But since you're cute, I'll give you another chance."],
			[111] = L["%%t%%, you have been weighed, and you have been measured... and I have been found rezing."],
			[112] = L["You don't deserve a cute rez macro, %%t%%. You deserve to die. But you already did, so, um... yeah."],
			[113] = L["%%t%%, I gotta have more cowbell!"],
			[114] = L["What do you get when you combine %%t%% with some mobs? I don't know, but it looks pretty messy here trying to scrap it off the floor."],
			[115] = L["%%t%%'s soul seems to have disconnected from the body. Researchers believe this may be directly related to HP levels. Please stand at least 10 yards away from the corpse while I attempt drastic measures to reconnect %%t%%."],
			[116] = L["1 mana = 1G, %%t%%. Your total comes up to 2517g, plus 211g tax. By accepting this rez, you hereby accept the life binding contract. Make sure you also read the EULA. Have a nice day!"],
			[117] = L["Priests, get Shackle ready. I've only done this twice, and one turned all 'Army of Darkness.'"],
			[118] = L["Eww... gross, %%t%%. My res is not an aoe spell! Would someone mind helping mop the pieces closer together so I don't bind %%t%%'s soul to just an arm or toe?"],
			[119] = L["Sorry %%t%%, it would have interrupted my healing rotation if I had healed you."],
			[120] = L["Alright, %%t%%, I'm going to res you. Just relax and let your mind go blank... shouldn't be hard for you, considering you're, well, dead."],
			[121] = L["So you just figured you'd come here, to the most hostile environment known to men, with no training of any kind, and see how it went? What was going through your head, %%t%%?"],
			[122] = L["Arise, %%t%%, and we will see if your insanity can be cured."],
			[123] = L["Nothing's over while I'm breathing, %%t%%!"],
			[124] = L["Sooner or later, though, %%t%%, you always have to wake up."],
			[125] = L["Sometimes your whole life boils down to one insane move... lucky for you, %%t%%, I can give you another chance."],
			[126] = L["I find your lack of life disturbing, %%t%%."],
			[127] = L["Not a bad bit of resurrection, huh, %%t%%? You know, sometimes I amaze even myself."],
			[128] = L["Uh, we had a slight weapons malfunction, but uh... %%t%%'s perfectly all right now. We're fine. We're all fine here now, thank you. How are you?"],
			[129] = L["You know, between the raid's howling and %%t%% blasting everything in sight, it's a wonder the whole instance doesn't know we're here."],
			[130] = L["You don't need to see %%t%%'s identification. This isn't the corpse you're looking for. You can go about your business. Move along."],
			[131] = L["I felt a great disturbance in the Force, as if millions of voices suddenly cried out in terror and were suddenly silenced. I fear something terrible has happened... oh wait, no, that was just %%t%%."],
			[132] = L["Judging by your current state, I suggest a new tactic, %%t%%: let the wookie win."],
			[133] = L["We're not gonna die. We can't die, %%t%%. You know why? Because we are so...very...pretty."],
			[134] = L["%%t%% died! What can we do? We're all doomed! Who's healing this party!? ...Oh right, that would be me. Back to work."],
			[135] = L["Here lies my beloved %%t%%, a lot less attractive now that you're all corpsified and gross."],
			[136] = L["Please note, %%t%%, that there is a consequence for failure. Any time your health bar reaches 0 percent will result in an 'unsatisfactory' mark on your official testing record, followed by death."],
			[137] = L["Greetings, %%t%%. In dangerous raiding environments, the Enrichment Center promises to always provide useful advice. For instance, having 0 percent health will kill you. Try to avoid it."],
			[138] = L["Uh oh. Somebody cut the cake. I told them to wait for you, %%t%%, but they cut it anyway. There is still some left, though, if you hurry and click accept."],
			[139] = L["Did you know, %%t%%, that grief counseling and cake will be available at the end of the testing period?"],
			[140] = L["The Enrichment Center is committed to the well-being of all participants. Cake and Grief Counseling will be available at the conclusion of the test. Thank you, %%t%%, for helping us help you help us all."],
			[141] = L["Area and State regulations do not allow %%t%% to remain here, alone and companionless."],
			[142] = L["Not in cruelty - Not in wrath - The REAPER came today - An ANGEL visited - %%t%%'s gray path - And took the cube away"],
			[143] = L["Because %%t%% could not stop for Death, He kindly stopped for me; the cube had food and maybe ammo. And immortality."],
			[144] = L["So I'm GLaD %%t%% got burned. Think of all the things we learned for the people who are still alive."],
			[145] = L["Aperture Science; We do what we must because we can. For the good of all of us. Except %%t%% who is dead."],
			[146] = L["Who's going to make the cake when %%t%% is gone? Me?"],
			[147] = L["This next boss could take a very, very long time. If you become lightheaded from thirst, feel free to pass out, like %%t%%. An intubation associate will be dispatched to revive you with peptic salve and adrenaline."],
			[148] = L["As part of a previously mentioned required test protocol, we can no longer lie to you. When the testing is over, %%t%% will be...*bzzt*... missed."],
			[149] = L["Despite the best efforts of the Enrichment Center Staff to ensure the safe performance of all authorized activities, %%t%% has managed to die inside this room. A complimentary revival will be performed in three, two, one..."],
			[150] = L["While safety is one of many Enrichment Center goals, the Aperture Science High-Energy Zones seen in many cases as black, green, red, or blue puddles can and have caused permanent disabilities, such as vaporization. Please be careful, %%t%%."],
			[151] = L["As part of a required raid protocol, a previous statement suggesting that we would not heal this raid was an outright fabrication. Good job, %%t%%. As part of a required raid protocol, we will stop enhancing the truth in three, two, o--"],
			[152] = L["%%t%%: please be advised that a noticeable taste of blood is not part of any protocol, but an unintended side effect of the Aperture Science Resurrection Grille, which may, in semi-rare cases, emancipate dental fillings, crowns, tooth enamel, and teeth."],
			[153] = L["Take heart %%t%%, for you have curried the favor of Boo, the only miniature giant space hamster in the Realm!"],
			[154] = L["This death will not dampen %%t%%'s wrath. %%t%% will be ressed! Butts shall be liberally kicked in good measure."],
			[155] = L["Will you accept my res, %%t%%? We must join together once more, and our fury will be such that bards will run their quills dry! Yes, ink will be scarce where e'er we go."],
			[156] = L["This dieing behavior must not continue, %%t%%. Feel the burning stare of my hamster and change your ways."],
			[157] = L["We must rest and get healthy. It would not do for us to be caught unsuspecting with %%t%% in this kind of shape."],
			[158] = L["Yes! Revenge! Revenge for precious %%t%%! Tremble, ye evil raid bosses, for the mighty Boo will have your eyes!! RAAAARRRRGHH!!"],
			[159] = L["%%t%%, there's something you better understand about me, 'cause it's important and one day your life may depend on it. I am definitely a madman with an acorn!"],
			[160] = L["The daleks would say that %%t%% is superior to them in only one respect: %%t%% is better at dyeing."],
			[161] = L["Such was the way %%t%% ended. Not with a bang, but a whimper."],
			[162] = L["It's our strongest impulse, to cling to life with every fiber of being! %%t%%'s only doing what everyone before us has tried to do. We've simply been more... successful at it."],
			[163] = L["The wonderful world of raiding... the prettier it looks, the more likely it is to kill you, eh %%t%%?"],
			[164] = L["You need to get yourself a better dictionary, %%t%%. When you do, look up 'raid wipe.' You'll find a little picture of me there, and the caption'll read 'Over my dead body.'"],
			[165] = L["Everybody knows that everybody dies, %%t%%. But not every day. Not today... every once in a very long while, every day in a million days, when the wind stands fair and the Doctor comes to call, everybody lives."],
			[166] = L["One lifetime's been too short for %%t%% to do everything they'd like. How much more they'll get done in two, or three, or four resses."]
		}
	}
}

-- standard methods ---------------------------------------------------------

function SmartRes2:OnInitialize()
	-- register saved variables with AceDB
	db = LibStub("AceDB-3.0"):New("SmartRes2DB", defaults, "Default")
	db.RegisterCallback(self, "OnProfileChanged", "OnProfileChanged")
	db.RegisterCallback(self, "OnProfileCopied", "OnProfileChanged")
	db.RegisterCallback(self, "OnProfileReset", "OnProfileChanged")
	defaults = nil -- done with the table, so get rid of it
	self.db = db
	local enabled = not self.db.profile.disableAddon
	self:SetEnabledState(enabled)

	-- addon options table
	local options = {
		name = "SmartRes2",
		handler = SmartRes2,
		type = "group",
		childGroups = "tab",
		args = {
			barsOptionsTab = {
				name = L["Res Bars"],
				desc = L["Options for the res bars"],
				type = "group",
				order = 1,
				args = {
					barsOptionsHeader = {
						order = 1,
						type = "header",
						name = L["Res Bars"]
					},
					hideAnchor = {
						order = 20,
						type = "toggle",
						name = L["Hide Anchor"],
						desc = L["Toggles the anchor for the res bars so you can move them"],
						get = function() return self.db.profile.hideAnchor end,
						set = function(info, value)
							self.db.profile.hideAnchor = value
							if value then
								self.res_bars:HideAnchor()
							else
								self.res_bars:ShowAnchor()
							end
						end
					},
					disableAddon = {
						order = 30,
						type = "toggle",
						name = L["Disable SmartRes2"],
						desc = L["Completely disable Smartres2"],
						get = function() return self.db.profile.disableAddon end,
						set = function(info, value)
							self.db.profile.disableAddon = value
							if value then
								self:Disable()
							else
								self:Enable()
							end
						end
					},					
					visibleResBars = { 
						order = 40,
						type = "toggle",
						name = L["Show Bars"],
						desc = L["Show or hide the res bars. Everything else will still work"],
						get = function() return self.db.profile.visibleResBars end,
						set = function(info, value) self.db.profile.visibleResBars = value end
					},					
					reverseGrowth = {
						order = 50,
						type = "toggle",
						name = L["Grow Upwards"],
						desc = L["Make the res bars grow up instead of down"],
						get = function() return self.db.profile.reverseGrowth end,
						set = function(info, value)
							self.db.profile.reverseGrowth = value
							self.res_bars:ReverseGrowth(value)
						end
					},					
					resBarsIcon = {
						order = 60,
						type = "toggle",
						name = L["Show Icon"],
						desc = L["Show or hide the icon for res spells"],
						get = function() return	self.db.profile.resBarsIcon end,
						set = function(info, value)
							self.db.profile.resBarsIcon = value
							if value then
								self.res_bars:ShowIcon()
							else
								self.res_bars:HideIcon()
							end
						end
					},					
					showBattleRes = {
						order = 70,
						type = "toggle",
						name = L["Show Battle Resses"],
						desc = L["Show bars for Rebirth"],
						get = function() return self.db.profile.showBattleRes end,
						set = function(info, value)	self.db.profile.showBattleRes = value end
					},					
					classColours = {
						order = 80,
						type = "toggle",
						name = L["Class Colours"],
						desc = L["Use class colours for the target on the res bars"],
						get = function() return self.db.profile.classColours end,
						set = function(info, value)	self.db.profile.classColours = value end
					},
					scale = {
						order = 90,
						type = "range",
						name = L["Scale"],
						desc = L["Set the scale for the res bars"],
						get = function() return self.db.profile.scale end,
						set = function(info, value)
							self.db.profile.scale = value
							self.res_bars:SetScale(value)
						end,
						min = 0.5,
						max = 2,
						step = 0.05
					},
					resBarsAlpha = {
						order = 100,
						type = "range",
						name = L["Alpha"],
						desc = L["Set the Alpha for the res bars"],
						get = function() return self.db.profile.resBarsAlpha end,
						set = function(info, value)
							self.db.profile.resBarsAlpha = value
							self.res_bars:SetAlpha(value)
						end,
						min = 0.1,
						max = 1,
						step = 0.1
					},
					resBarsTexture = {
						order = 110,
						type = "select",
						dialogControl = "LSM30_Statusbar",
						name = L["Texture"],
						desc = L["Select the texture for the res bars"],
						values = AceGUIWidgetLSMlists.statusbar,
						get = function() return self.db.profile.resBarsTexture end,
						set = function(info, value)	self.db.profile.resBarsTexture = value end
					},
					resBarsBorder = {
						order = 120,
						type = "select",
						dialogControl = "LSM30_Border",
						name = L["Border"],
						desc = L["Select the border for the res bars"],
						values = AceGUIWidgetLSMlists.border,
						get = function() return self.db.profile.resBarsBorder end,
						set = function(info, value) self.db.profile.resBarsBorder = value end
					},
					borderThickness = {
						order = 130,
						type = "range",
						name = L["Border Thickness"],
						desc = L["Set the thickness of the res bars border"],
						get = function() return self.db.profile.borderThickness end,
						set = function(info, value) self.db.profile.borderThickness = value end,
						min = 1,
						max = 10,
						step = 1
					},
					horizontalOrientation = {
						order = 140,
						type = "select",
						name = L["Horizontal Direction"],
						desc = L["Change the horizontal direction of the res bars"],
						values = {
							["LEFT"] = L["Right to Left"],
							["RIGHT"] = L["Left to Right"], 
						},
						get = function() return self.db.profile.horizontalOrientation end,
						set = function(info, value) self.db.profile.horizontalOrientation = value end
					},
					resBarsColour = {
						order = 150,
						type = "color",
						name = L["Bar Colour"],
						desc = L["Pick the colour for non-collision (not a duplicate) res bar"],
						hasAlpha = true,
						get = function()
							local t = self.db.profile.resBarsColour
							return t.r, t.g, t.b, t.a
						end,
						set = function(info, r, g, b, a)
							local t = self.db.profile.resBarsColour
							t.r, t.g, t.b, t.a = r, g, b, a
						end
					},
					collisionBarsColour = {
						order = 160,
						type = "color",
						name = L["Duplicate Bar Colour"],
						desc = L["Pick the colour for collision (duplicate) res bars"],
						hasAlpha = true,
						get = function()
							local t = self.db.profile.collisionBarsColour
							return t.r, t.g, t.b, t.a
						end,
						set = function(info, r, g, b, a)
							local t = self.db.profile.collisionBarsColour
							t.r, t.g, t.b, t.a = r, g, b, a
						end
					},
					waitingBarsColour = {
						order = 170,
						type = "color",
						name = L["Waiting Bar Colour"],
						desc = L["Pick the colour for collision (player waiting for accept) res bars"],
						hasAlpha = true,
						get = function()
							local t = self.db.profile.waitingBarsColour
							return t.r, t.g, t.b, t.a
						end,
						set = function(info, r, g, b, a)
							local t = self.db.profile.waitingBarsColour
							t.r, t.g, t.b, t.a = r, g, b, a
						end
					},
					resBarsTestBars = {
						order = 200,
						type = "execute",
						name = L["Test Bars"],
						desc = L["Show the test bars"],
						func = function() self:StartTestBars() end
					},
				}
			},
			resChatTextTab = {
				name = L["Chat Output"],
				desc = L["Chat output options"],
				type = "group",
				order = 2,
				args = {
					resChatHeader = {
						order = 1,
						type = "header",
						name = L["Chat Output"]
					},
					randMsgs = {
						order = 2,
						type = "toggle",
						name = L["Random Res Messages"],
						desc = L["Turn random res messages on or keep the same message. Default is off"],
						get = function() return self.db.profile.randMsgs end,
						set = function(info, value)	self.db.profile.randMsgs = value end
					},
					chatOutput = {
						order = 3,
						type = "select",
						name = L["Chat Output Type"],
						desc = L["Where to print the res message. Raid, Party, Say, Yell, Guild, smart Group, or None"],
						values = {
							["0-none"] = L["None"],
							group = L["Group"],
							guild = L["Guild"],
							party = L["Party"],
							raid = L["Raid"],
							say = L["Say"],
							whisper = L["Whisper"],
							yell = L["Yell"],							
						},
						get = function() return self.db.profile.chatOutput end,
						set = function(info, value)	self.db.profile.chatOutput = value end
					},
					notifySelf = {
						order = 4,
						type = "toggle",
						name = L["Self Notification"],
						desc = L["Prints a message to yourself whom you are ressing"],
						get = function() return self.db.profile.notifySelf end,
						set = function(info, value)	self.db.profile.notifySelf = value end
					},
					notifyCollision = {
						order = 5,
						type = "select",
						name = L["Duplicate Res Targets"],
						desc = L["Notify a resser they created a collision. Could get very spammy"],
						values = {
							["0-off"] = L["Off"],
							group = L["Group"],
							guild = L["Guild"],
							party = L["Party"],
							raid = L["Raid"],
							say = L["Say"],
							whisper = L["Whisper"],
							yell = L["Yell"],
						},
						get = function() return self.db.profile.notifyCollision end,
						set = function(info, value)	self.db.profile.notifyCollision = value	end
					},
					customMessage = {
						order = 6,
						type = "input",
						name = L["Custom Message"],
						desc = L["Your message.  Use 'me' for yourself and 'you' for target"],
						get = function() return self.db.profile.customchatmsg end,
						set = function(info, value) self:AddCustomMsg(value) end,
					}
				}
			},
			keyBindingsTab = {
				name = L["Key Bindings"],
				desc = L["Set the keybindings"],
				type = "group",
				order = 3,
				args = {
					autoResKey = {
						order = 1,
						type = "keybinding",
						name = L["Auto Res Key"],
						desc = L["For ressing targets who have not released their ghosts"],
						get = function() return self.db.profile.autoResKey end,
						set = function(info, value)	self.db.profile.autoResKey = value	end
					},
					manualResKey = {
						order = 2,
						type = "keybinding",
						name = L["Manual Target Key"],
						desc = L["Gives you the pointer to click on corpses"],
						get = function() return self.db.profile.manualResKey end,
						set  = function(info, value) self.db.profile.manualResKey = value end
					},
					castCommand = {
						order = 3,
						type = "description",
						name = L["The command \"cast\" will fire the smart Resurrection function. Usage: /sr cast or /smartres cast. Not necessary if you use the auto res key"]
					}
				}
			},
			creditsTab = {
				name = L["SmartRes2 Credits"],
				desc = L["About the author and SmartRes2"],
				type = "group",
				order = 6,
				args = {
					creditsHeader1 = {
						order = 1,
						type = "header",
						name = L["Credits"]
					},
					creditsDesc1 = {
						order = 2,
						type = "description",
						name = "* "..L["Massive kudos to Maia, Kyahx, and Poull for the original SmartRes. SmartRes2 was largely possible because of DathRarhek's LibResComm-1.0 so a big thanks to him."]
					},
					creditsDesc2 = {
						order = 3,
						type = "description",
						name = "* "..L["I would personally like to thank Jerry on the wowace forums for coding the new, smarter, resurrection function."]
					},
					creditsDesc3 = {
						order = 4,
						type = "description",
						name = "* "..L["Many bugfixes and development help from Onaforeignshore"]
					},
					creditsDesc5 = {
						order = 5,
						type = "description",
						name = "* "..L["German translation by Farook, Black_Mystics, Xevilgrin, and Dessa"],
					},
					creditsDesc6 = {
						order = 6,
						type = "description",
						name = "* "..L["French translation by Ckeurk and Xilbar"],
					},
					creditsDesc7 = {
						order = 7,
						type = "description",
						name = "* "..L["Latin American Spanish and Spanish translation by Silmano"],
					},
					creditsDesc8 = {
						order = 8,
						type = "description",
						name = "* "..L["Russian translation by Xenobios"],
					},
					creditsDesc9 = {
						order = 9,
						type = "description",
						name = "* "..L["Traditional Chinese translation by Whocare"],
					},
					creditsDesc10 = {
						order = 10,
						type = "description",
						name = "* "..L["Additional translations by Mattbnr"]
					}
				}
			}
		}
	}
	-- add the 'Profiles' section
	options.args.profilesTab = LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db)
	options.args.profilesTab.order = 5

	-- Register your options with AceConfigRegistry
	LibStub("AceConfig-3.0"):RegisterOptionsTable("SmartRes2", options)

	-- Add your options to the Blizz options window using AceConfigDialog
	self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("SmartRes2", "SmartRes2")	

	-- support for LibAboutPanel
	if LibStub:GetLibrary("LibAboutPanel", true) then
		self.optionsFrame[L["About"]] = LibStub("LibAboutPanel").new("SmartRes2", "SmartRes2")
	end

	-- add console commands
	self:RegisterChatCommand("sr", "SlashHandler")
	self:RegisterChatCommand("smartres", "SlashHandler")	

	-- prepare spells
	local resSpells = { -- getting the spell names
		PRIEST = GetSpellInfo(2006), -- Resurrection
		SHAMAN = GetSpellInfo(2008), -- Ancestral Spirit
		DRUID = GetSpellInfo(50769), -- Revive
		PALADIN = GetSpellInfo(7328) -- Redemption
	}
	self.resSpellIcons = { -- need the icons too, for the res bars
		PRIEST = select(3, GetSpellInfo(2006)), 	-- Resurrection
		SHAMAN = select(3, GetSpellInfo(2008)), 	-- Ancestral Spirit
		DRUID = select(3, GetSpellInfo(50769)), 	-- Revive
		PALADIN = select(3, GetSpellInfo(7328)) 	-- Redemption
	}  
	self.playerClass = select(2, UnitClass("player"))
	self.playerSpell = resSpells[self.playerClass]
	
	-- create DataBroker Launcher
	if DataBroker then
		local launcher = DataBroker:NewDataObject("SmartRes2", {
		type = "launcher",
		icon = self.resSpellIcons[self.playerClass] or self.resSpellIcons.PRIEST,
		OnClick = function(clickedframe, button)
				  if button == "LeftButton" then
					  -- keep our options table in sync with the ldb object state
					  self.db.profile.hideAnchor = not self.db.profile.hideAnchor
					  if self.db.profile.hideAnchor then
						  self.res_bars:HideAnchor()
					  else
						  self.res_bars:ShowAnchor()
					  end
					  LibStub("AceConfigRegistry-3.0"):NotifyChange("SmartRes2")
				  elseif button == "RightButton" then
					  _G.InterfaceOptionsFrame_OpenToCategory(self.optionsFrame)
				  end
		end,
		OnTooltipShow = function(self)
					GameTooltip:AddLine("SmartRes2".." "..version, HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
					GameTooltip:AddLine(L["Left click to lock/unlock the res bars. Right click for configuration."], NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b)
					GameTooltip:Show()
				end
	})
		self.launcher = launcher
	end	

	-- create a secure button for ressing
	local resButton = CreateFrame("button", "SmartRes2Button", UIParent, "SecureActionButtonTemplate")
	resButton:SetAttribute("type", "spell")
	resButton:SetScript("PreClick", function() self:Resurrection() end)
	self.resButton = resButton

	-- create the Res Bars and set the user preferences
	self.res_bars = self:NewBarGroup("SmartRes2", self.db.horizontalOrientation, 300, 15, "SmartRes2_ResBars")
	self.res_bars:SetPoint("CENTER", UIParent, "CENTER", self.db.profile.resBarsX, self.db.profile.resBarsY)
	self.res_bars:SetUserPlaced(true)
	if self.db.profile.hideAnchor then
		self.res_bars:HideAnchor()
	else
		self.res_bars:ShowAnchor()
	end
end

function SmartRes2:OnEnable()
	-- called when SmartRes2 is enabled
	self:RegisterEvent("PLAYER_REGEN_DISABLED")
	self:RegisterEvent("PLAYER_REGEN_ENABLED")
	self:RegisterEvent("RAID_ROSTER_UPDATE")	
	Media.RegisterCallback(self, "OnValueChanged", "UpdateMedia")
	ResComm.RegisterCallback(self, "ResComm_ResStart")
	ResComm.RegisterCallback(self, "ResComm_ResEnd")
	ResComm.RegisterCallback(self, "ResComm_Ressed")
	ResComm.RegisterCallback(self, "ResComm_ResExpired")
	self.res_bars.RegisterCallback(self, "FadeFinished")
	self.res_bars.RegisterCallback(self, "AnchorMoved", "ResAnchorMoved")
	self:BindKeys()
end

function SmartRes2:OnDisable()
	-- called when SmartRes2 is disabled
end

-- process slash commands ---------------------------------------------------
function SmartRes2:SlashHandler(input)
	input = input:lower()

	if input == "cast" then
		self:Resurrection()
	else
		_G.InterfaceOptionsFrame_OpenToCategory(self.optionsFrame)
	end
end

-- enable and disable SmartRes2 ---------------------------------------------
-- using the power of Ace3 to reduce code -----------------------------------

-- enable SmartRes2 ----------------------------------------------------------
function SmartRes2:Enable()
end

-- disable SmartRes2 completely ----------------------------------------------
function SmartRes2:Disable()
	self:UnBindKeys()
end

-- General callback functions -----------------------------------------------

-- called when user changes profile
function SmartRes2:OnProfileChanged()
	db = self.db
end

-- called when user changes the texture of the bars
function SmartRes2:UpdateMedia(callback, type, handle)
	if type == "statusbar" then
		self.res_bars:SetTexture(Media:Fetch("statusbar", self.db.profile.resBarsTexture))
	elseif type == "border" then
		self.res_bars:SetBackdrop({
			edgeFile = Media:Fetch("border", self.db.profile.resBarsBorder),
			tile = false,
			tileSize = self.db.profile.scale + 1,
			edgeSize = self.db.profile.borderThickness,
			insets = { left = 0, right = 0, top = 0, bottom = 0 }
		})
	end
end

function SmartRes2:AddCustomMsg(msg)
	msg = string.gsub(msg, "me", "%%%%p%%%%")
	msg = string.gsub(msg, "you", "%%%%t%%%%")
	self.db.profile.customchatmsg = msg
end

-- ResComm library callback functions ---------------------------------------

-- ResComm events - called when res is started
function SmartRes2:ResComm_ResStart(event, sender, endTime, targetName)
	-- check if we have the person in our table yet, and if not, add them
	--self:Print("ResStart - "..sender.." ressing "..targetName)
	if doingRessing[sender] then
		return
	end
	doingRessing[sender] = {
		endTime = endTime,
		target = targetName
	}
	self:CreateResBar(sender)

	if waitingForAccept[targetName] then self:AddWaitingBars(sender, targetName) end

	local oldsender = self:CheckResTarget(targetName, sender) 

	if oldsender then		--target already being ressed
		self:AddCollisionBars(sender, targetName, oldsender)
	end
	local isSame = UnitIsUnit(sender, "player")

	if isSame == 1 then -- make sure only the player is sending messages
		local channel = self.db.profile.chatOutput:upper()

		if channel == "GROUP" then
			if UnitInRaid("player") then
				channel = "RAID"
			elseif GetNumPartyMembers() > 0 then
				channel = "PARTY"
			end
		end

		if channel ~= "0-NONE" then -- if it is "none" then don't send any chat messages
			local msg = L["%%p%% is ressing %%t%%"]

			if self.db.profile.randMsgs then
				msg = self.db.profile.randChatTbl[math.random(#self.db.profile.randChatTbl)]
			elseif self.db.profile.customchatmsg ~= "" then
				msg = self.db.profile.customchatmsg
			end
			msg = string.gsub(msg, "%%%%p%%%%", sender)
			msg = string.gsub(msg, "%%%%t%%%%", targetName)

			SendChatMessage(msg, channel, nil, (channel == "WHISPER") and targetName or nil)
		end

		if self.db.profile.notifySelf then
			self:Print((L["You are ressing %s"]):format(targetName))
		end
	end
end

function SmartRes2:CheckResTarget(target, newsender)
	for sender, info in pairs(doingRessing) do
		if info.target == target and sender ~= newsender then return sender end
	end
	return nil
end

-- ResComm events - called when res ends
function SmartRes2:ResComm_ResEnd(event, sender, target)
	-- did the cast fail or complete?
	if not doingRessing[sender] then return end
	self:DeleteResBar(sender)
	-- add the target to our waiting list, and save who the last person to res him was
	waitingForAccept[target] = { target = target, sender = sender, endTime = doingRessing[sender].endTime }
	doingRessing[sender] = nil
	if self.db.profile.visibleResBars then 
		local oldsender = self:CheckResTarget(target, sender) 
		if oldsender and not self:CheckResTarget(target, oldsender) then	--collision bar existed and only 1 exists
			self:DeleteCollisionBars(sender, target, oldsender)
		end
	end
end

-- ResComm events - called when player accepts res
function SmartRes2:ResComm_Ressed(event, target)
	-- target accepted, remove from list
	waitingForAccept[target] = nil
end

-- ResComm events - called when res box disappears or player declines res
function SmartRes2:ResComm_ResExpired(event, target)
	-- target declined, remove from list
	waitingForAccept[target] = nil
end

-- ResComm events - called when bar finished fading
function SmartRes2:FadeFinished(event, bar, name)
	self.res_bars:ReleaseBar(bar)
end

-- Blizzard callback functions ----------------------------------------------

-- Important Note: Since the release of patch 3.2, certain fights in the game fire the 
-- "PLAYER_REGEN_DISABLED" event continuously during combat causing any subsequent events
-- we might trigger as a result to also fire continuously. It is recommended therefore to
-- use a checking variable that is set to 'on/1/etc' when entering combat and back to
-- 'off/0/etc' only when exiting combat and then use this as the final determinant on
-- whether or not to action a subsequent event.
function SmartRes2:PLAYER_REGEN_DISABLED()
	-- don't confuse the variable below with being in combat - we use it to see if we've run
	-- the code below on entry into combat already so that we only run it once
	if not in_combat then
		self:UnBindKeys()
		-- disable callbacks during battle if we don't want to see battle resses
		if not self.db.profile.showBattleRes then
			ResComm.UnregisterAllCallbacks(self)
		end
		-- clear the ressing tables
		wipe(doingRessing)
		wipe(waitingForAccept)
		wipe(resBars)
		LastRes = nil
	end
	in_combat = true
end

function SmartRes2:PLAYER_REGEN_ENABLED()
	self:BindKeys()
	-- reenable callbacks during battle if we don't want to see battle resses
	if not self.db.profile.showBattleRes then
		ResComm.RegisterCallback(self, "ResComm_ResStart")
		ResComm.RegisterCallback(self, "ResComm_ResEnd")
		ResComm.RegisterCallback(self, "ResComm_Ressed")
		ResComm.RegisterCallback(self, "ResComm_ResExpired")
		self.res_bars.RegisterCallback(self, "FadeFinished")
	end
	in_combat = false
end

-- key binding functions ----------------------------------------------------

function SmartRes2:BindKeys()
	-- only binds keys if the player can cast an out of combat res spell
	if not self.playerSpell then return end
	SetOverrideBindingClick(self.resButton, false, self.db.profile.autoResKey, "SmartRes2Button")
	SetOverrideBindingSpell(self.resButton, false, self.db.profile.manualResKey, self.playerSpell)
end

function SmartRes2:UnBindKeys()
	ClearOverrideBindings(self.resButton)
end

-- anchor management functions ----------------------------------------------

function SmartRes2:ResAnchorMoved(_, _, x, y)
	self.db.profile.resBarsX, self.db.profile.resBarsY = x, y
end

-- smart ressurection determination functions -------------------------------
local raidUpdated

function SmartRes2:RAID_ROSTER_UPDATE()
	raidUpdated = true
end

local unitOutOfRange, unitBeingRessed, unitDead, unitWaiting, unitGhost
local SortedResList = {}
local CLASS_PRIORITIES = {
	-- There might be 10 classes, but SHAMANs and DRUIDs res at equal efficiency, so no preference
	-- non healers who use Mana should be followed after healers, as they are usually buffers
	-- or pet summoners (ie: Mana burners)
	-- res non Mana users last (HUNTERs will get moved once Catalclysm hits)
	PRIEST = 1, 
	PALADIN = 2, 
	SHAMAN = 3, 
	DRUID = 3, 
	MAGE = 4, 
	WARLOCK = 4, 
	HUNTER = 4,
	DEATHKNIGHT = 5,
	ROGUE = 5,
	WARRIOR = 5
}

-- create resurrection tables
local function getClassOrder(unit)
	local _, c = UnitClass(unit)
	local lvl = UnitLevel(unit)
	return CLASS_PRIORITIES[c], lvl
end

local function verifyUnit(unit)
	-- unit is the next candidate. there is NO way to check LoS, so don't ask!
	if UnitIsGhost(unit) then
		unitGhost = true
		unitDead = true
		return nil
	end
	if not UnitIsDead(unit) then return nil end
	unitDead = true
	if unit == LastRes then return nil end
	if ResComm:IsUnitBeingRessed(unit) then unitBeingRessed = true return nil end
	if waitingForAccept[unit] then unitWaiting = true return nil end
	if IsSpellInRange(SmartRes2.playerSpell, unit) ~= 1 then unitOutOfRange = true return nil end
	return true
end

--sort function only called when raid has actually changed (avoided looking up unit names/classes everytime we click the res button)
local function SortCurrentRaiders()
	local num = GetNumRaidMembers()
	local member = "raid"
	if num == 0 then
		num = GetNumPartyMembers()
		member = "party"
	end
	wipe(SortedResList)
	for i = 1, num do
		local id = member .. i
		local name = UnitName(id)
		local resprio, lvl = getClassOrder(name)
		tinsert(SortedResList, {name = name, resprio = resprio, level = lvl})
	end
	tsort(SortedResList, function(a,b) 
		if a.resprio == b.resprio then
			return a.level > b.level
		else 
			return a.resprio < b.resprio
		end
	end)
	raidUpdated = nil
end


local function getBestCandidate()
	unitOutOfRange, unitBeingRessed, unitDead, unitWaiting, unitGhost = nil, nil, nil, nil, nil
	if raidUpdated then SortCurrentRaiders() end	--only resort if raid changed	
	for _, data in ipairs(SortedResList) do
		local unit = data.name
		local validUnit = verifyUnit(unit)
		if validUnit then
			return unit
		end
	end
	return nil
end

function SmartRes2:Resurrection()
	local resButton = self.resButton
	resButton:SetAttribute("unit", nil)

	if GetNumPartyMembers() == 0 and not UnitInRaid("player") then
		self:Print(L["You are not in a group."])
		return
	else
		resButton:SetAttribute("spell", self.playerSpell)
	end

	-- check if the player has enough Mana to cast a res spell. if not, no point in continuing. same if player is not a sender 
	local isUsable, outOfMana = IsUsableSpell(self.playerSpell) 
	if outOfMana == 1 then 
	   self:Print(L["You don't have enough Mana to cast a res spell."]) 
	   return 
	elseif isUsable ~= 1 then 
		self:Print(L["You cannot cast res spells."]) -- in the final code, you should never see this message
		return 
	end

	local unit = getBestCandidate()
	if unit then
		resButton:SetAttribute("unit", unit)		
		LastRes = unit
	else
		if unitOutOfRange then
			self:Print(L["There are no bodies in range to res."])
		elseif unitBeingRessed or unitWaiting then
			self:Print(L["All dead units are being ressed."])
		elseif not unitDead then
			self:Print(L["Everybody is alive. Congratulations!"])
			wipe(waitingForAccept)
		elseif unitGhost then
			self:Print(L["All dead units have released."])
		end
	end
end

-- resbar functions ---------------------------------------------------------

local function ClassColouredName(name)
	if not name then return "|cffcccccc"..L["Unknown"].."|r" end
	local _, class = UnitClass(name)
	if not class then return "|cffcccccc"..name.."|r" end
	local c = RAID_CLASS_COLORS[class]
	return ("|cff%02X%02X%02X%s|r"):format(c.r * 255, c.g * 255, c.b * 255, name)
end

function SmartRes2:CreateResBar(sender)
	if not self.db.profile.visibleResBars then return end
	local text
	local senderClass = select(2, UnitClass(sender))
	if senderClass == "DRUID" and in_combat then
		icon = select(3, GetSpellInfo(20484))
	else		
		icon = self.resSpellIcons[select(2, UnitClass(sender))] or self.resSpellIcons.PRIEST
	end
	local info = doingRessing[sender]
	local time = info.endTime - GetTime()
	
	if self.db.profile.classColours then
		text = (L["%s is ressing %s"]):format(ClassColouredName(sender), ClassColouredName(info.target))
	else
		text = (L["%s is ressing %s"]):format(sender, info.target)
	end

	-- args are as follows: lib:NewTimerBar(name, text, time, maxTime, icon, flashTrigger)
	local bar = self.res_bars:NewTimerBar(sender, text, time, nil, icon, 0)
	local t = self.db.profile.resBarsColour
	bar:SetBackgroundColor(t.r, t.g, t.b, t.a)
	bar:SetColorAt(0, 0, 0, 0, 1) -- set bars to be black behind the cast bars
	orientation = (self.db.profile.horizontalOrientation == "RIGHT") and Bars.RIGHT_TO_LEFT or Bars.LEFT_TO_RIGHT
	bar:SetOrientation(orientation)
	bar:SetTexture(Media:Fetch("statusbar", self.db.profile.resBarsTexture))
	bar:SetBackdrop({
		edgeFile = Media:Fetch("border", self.db.profile.resBarsBorder),
		tile = false,
		tileSize = self.db.profile.scale + 1,
		edgeSize = self.db.profile.borderThickness,
		insets = { left = 0, right = 0, top = 0, bottom = 0 }
	})
	resBars[sender] = bar
end

function SmartRes2:DeleteResBar(sender)
	local info = doingRessing[sender]
	if not info then return end
	resBars[sender]:Fade(0.1)
	resBars[sender] = nil
end

function SmartRes2:AddCollisionBars(sender, target, collisionsender)
	if self.db.profile.visibleResBars then 
		local t = self.db.profile.collisionBarsColour
		resBars[sender]:SetBackgroundColor(t.r, t.g, t.b, t.a)
	end
	local chatType = self:GetChatType()
	if chatType ~= "0-OFF" and sender ~= Player then
		SendChatMessage((L["SmartRes2 would like you to know that %s is already being ressed by %s."]):format(target, collisionsender), chatType, nil, sender)
	end
end

function SmartRes2:AddWaitingBars(sender, target)
	if self.db.profile.visibleResBars then 
		local t = self.db.profile.waitingBarsColour
		resBars[sender]:SetBackgroundColor(t.r, t.g, t.b, t.a)
	end
end

function SmartRes2:DeleteCollisionBars(sender, target, collisionsender)
	local t = self.db.profile.resBarsColour
	resBars[collisionsender]:SetBackgroundColor(t.r, t.g, t.b, t.a)
end

function SmartRes2:GetChatType()
	local chatType = self.db.profile.notifyCollision:upper()
	if chatType == "GROUP" then
		if GetNumRaidMembers() > 0 then
			chatType = "RAID"
		elseif GetNumPartyMembers() > 0 then
			chatType = "PARTY"
		end
	end
	return chatType
end

function SmartRes2:StartTestBars()
	-- we don't want the test bars to throw an error if notify collision is on
	local settings = self.db.profile.notifyCollision
	if settings ~= "0-off" then
		self.db.profile.notifyCollision = "0-off"
	end

	-- set up the test bars
	waitingForAccept["Someone"] = { target = "Someone", sender = "Dummy", endTime = GetTime() - 6 }
	self:ResComm_ResStart(nil, "Nursenancy", GetTime() + 4, "Frankthetank")
	self:ResComm_ResStart(nil, "Dummy", GetTime() + 8, "Frankthetank")
	self:ResComm_ResStart(nil, "Gabriel", GetTime() + 6, "Someone")
		
	-- clean up
	doingRessing["Nursenancy"] = nil
	doingRessing["Dummy"] = nil
	doingRessing["Gabriel"] = nil
	waitingForAccept["Someone"] = nil
	
	-- set the collision back to user preferences
	self.db.profile.notifyCollision = settings
end